#include "MyDeRBlock.h"
#include "BlockMwITransform.h"

#include <math.h>
#include <stdio.h>
#include <string>

#include "MyDefines.h"
#include myPATH(MY_INCLUDES_DIR,AuxCompFunctions.h)
#include myPATH(MY_INCLUDES_DIR,MTFWithIntegers.h)
#include myPATH(MY_INCLUDES_DIR,MwI.h)

#pragma warning (disable : 4996)




int MyDeRBlock::Init(const char* FName)
{
	FILE* f;
	f = fopen(FName, "rt");

	if (f == NULL)
		return 0;


	fscanf(f, "%d %d\n", &W, &H);
	fscanf(f, "%d \n", &BSize);
	fscanf(f, "%d \n", &MwIDelta);
	fscanf(f, "%lf \n", &IDWDistance);

	int cVectorSize;
	fscanf(f, "%d\n", &cVectorSize);
	ControlVector.clear();
	int a;
	for (int i = 0; i < cVectorSize; i++)
	{
		fscanf(f, "%d ", &a);
		ControlVector.push_back(a);
	}

	P = AllocateMatrix(W, H);
	for (int j = 0; j < H; j++)
	{
		for (int i = 0; i < W; i++)
		{
			fscanf(f, "%d ", &a);
			P[i][j] = a;
		}
		fscanf(f, "\n");
	}
	return 1;
}

void MyDeRBlock::ReconstructLUBorder()
{

	R[0][0] = P[0][0];

	for (int x = 1; x < W; x++)
		R[x][0] = UnWrapePositiveValues(P[x][0]) + R[x - 1][0];

	for (int y = 1; y < H; y++)
		R[0][y] = UnWrapePositiveValues(P[0][y]) + R[0][y - 1];
}

void MyDeRBlock::SetBlockCoordinates(int x, int y)
{
	if (x == 0)
		x1 = x + 1;
	else
		x1 = x;

	if (y == 0)
		y1 = y + 1;
	else
		y1 = y;

	x2 = x + BSize;
	if (x2 > W)
		x2 = W;

	y2 = y + BSize;
	if (y2 > H)
		y2 = H;

	dx = x2 - x1;
	dy = y2 - y1;

}


vector<int> MyDeRBlock::FillBlockBorderValues()
{
	vector<int> V;

	int x = x1 - 1;
	int y = y1 - 1;

	V.push_back(R[x][y]);

	for (int i = 0; i < dx + dy; i++)
	{
		x++;
		if (x < x2)
			V.push_back(R[x][y1 - 1]);

		y++;
		if (y < y2)
			V.push_back(R[x1 - 1][y]);
	}

	return V;
}

/*
void MyDeRBlock::ReconstructRawBlock(int x, int y, int minV)
{
	SetBlockCoordinates(x, y);
	for (int i = x1; i < x2; i++)
		for (int j = y1; j < y2; j++)
			R[i][j] = P[i][j] + minV;
}*/

void MyDeRBlock::ReconstructRawBlock(int x, int y)
{
/*	SetBlockCoordinates(x, y);
	for (int i = x1; i < x2; i++)
		for (int j = y1; j < y2; j++)
			R[i][j] = P[i][j] + minV;
			*/

	SetBlockCoordinates(x, y);
	int leftup = R[x1 - 1][y1 - 1];

	for (int i = x1; i < x2; i++)
		for (int j = y1; j < y2; j++)
			R[i][j] = UnWrapePositiveValues(P[i][j]) + leftup;
}



int MyDeRBlock::GetPredictionJPEGLSDe(int x, int y)
{
	if (x == 0)
		return R[x][y - 1];

	int a = R[x][y - 1];
	int c = R[x - 1][y - 1];
	int b = R[x - 1][y];

	if ((c > a) && (c > b))
		if (a < b)
			return a;
		else
			return b;

	if ((c < a) && (c < b))
		if (a > b)
			return a;
		else
			return b;

	int d = a + b - c;
	return d;
}

void MyDeRBlock::ReconstructJPEGLS(int x, int y)
{

	SetBlockCoordinates(x, y);
	for (int i = x1; i < x2; i++)
		for (int j = y1; j < y2; j++)
			R[i][j] = GetPredictionJPEGLSDe(i, j) - UnWrapePositiveValues(P[i][j]);
}

void MyDeRBlock::ReconstructHorizontal(int x, int y)
{
	SetBlockCoordinates(x, y);
	for (int i = x1; i < x2; i++)
		for (int j = y1; j < y2; j++)
			R[i][j] = R[i][j - 1] - UnWrapePositiveValues(P[i][j]);
}

void MyDeRBlock::ReconstructVertical(int x, int y)
{
	SetBlockCoordinates(x, y);
	for (int i = x1; i < x2; i++)
		for (int j = y1; j < y2; j++)
			R[i][j] = R[i - 1][j] - UnWrapePositiveValues(P[i][j]);
}


void MyDeRBlock::ReconstructMTF(int x, int y)
{
	SetBlockCoordinates(x, y);
	vector<int> BV = FillBlockBorderValues();

	vector<int> ImtfVector;
	for (int j = y1; j < y2; j++)
		for (int i = x1; i < x2; i++)
			ImtfVector.push_back(P[i][j]);

	MTFWithIntegers DeMtf;
	DeMtf.Init(0, 255, BV);
	vector<int> V = DeMtf.DoDeMTF(ImtfVector);
	int k = 0;

	for (int j = y1; j < y2; j++)
		for (int i = x1; i < x2; i++)
			R[i][j] = V[k++];
}



void MyDeRBlock::ReconstructMwI(int x, int y)
{
	SetBlockCoordinates(x, y);
	BlockMwITransform MwIT;
	MwIT.Init(0, 255, MwIDelta);
	int a, b, c, d;


	for (int j = y1; j < y2; j++)
		for (int i = x1; i < x2; i++)
		{
			a = R[i - 1][j];
			b = R[i - 1][j - 1];
			c = R[i][j - 1];
			d = -1;
			if (i + 1 < x2)
				d = R[i + 1][j - 1];

			R[i][j] = MwIT.MwIDe(a, b, c, d, P[i][j]);
		}
}


void MyDeRBlock::ReconstructDiagonal(int x, int y)
{
	SetBlockCoordinates(x, y);

	int xd, yd;
	for (int x = x1; x < x2; x++)
	{
		xd = x;
		yd = y1;
		while ((xd < x2) && (yd < y2))
		{
			R[xd][yd] = R[xd - 1][yd - 1] - UnWrapePositiveValues(P[xd][yd]);
			xd++;
			yd++;
		}
	}

	for (int y = y1 + 1; y < y2; y++)
	{
		xd = x1;
		yd = y;
		while ((xd < x2) && (yd < y2))
		{
			R[xd][yd] = R[xd - 1][yd - 1] - UnWrapePositiveValues(P[xd][yd]);
			xd++;
			yd++;
		}
	}
}


void MyDeRBlock::ReconstructIDW(int x, int y)
{
	vector<int> X, Y, V;
	SetBlockCoordinates(x, y);

	int xx = x1 - 1;
	int yy = y1 - 1;
	for (int i = xx; i < x2; i++)
	{
		X.push_back(i);
		Y.push_back(yy);
		V.push_back(R[i][yy]);
	}

	for (int j = yy; j < y2; j++)
	{
		X.push_back(xx);
		Y.push_back(j);
		V.push_back(R[xx][j]);
	}


	int v;
	for (int j = y1; j < y2; j++)
	{
		for (int i = x1; i < x2; i++)
		{
			v = ReturnIDW(X, Y, V, i, j, IDWDistance);
			R[i][j] = v - UnWrapePositiveValues(P[i][j]);

			X.push_back(i);
			Y.push_back(j);
			V.push_back(R[i][j]);
		}
	}


}

void MyDeRBlock::ReconstructUpLeftAverage(int x, int y)
{
	SetBlockCoordinates(x, y);

	for (int i = x1; i < x2; i++)
		for (int j = y1; j < y2; j++)
			R[i][j] = (int)(0.5 * (R[i - 1][j] + R[i][j - 1])) - UnWrapePositiveValues(P[i][j]);
}




int MyDeRBlock::GetPaethPrediction(int i, int j)
// a: left, b: up, c: left up pixel
{
	int a = R[i - 1][j];
	int b = R[i][j - 1];
	int c = R[i - 1][j - 1];

	int p = a + b - c;
	int pa = abs(p - a);
	int pb = abs(p - b);
	int pc = abs(p - c);
	if ((pa <= pb) && (pa <= pc)) return a;
	if (pb <= pc) return b;
	return c;
}

void MyDeRBlock::ReconstructPeath(int x, int y)
{
	SetBlockCoordinates(x, y);

	for (int j = y1; j < y2; j++)
		for (int i = x1; i < x2; i++)
			R[i][j] = GetPaethPrediction(i,j) - UnWrapePositiveValues(P[i][j]);
}

int MyDeRBlock::GetPredictionCALICDe(int x, int y)
{

	if ((x == 0) && (y == 0))
		return R[0][0];

	if (y == 0)
		return R[x][0] - R[x - 1][0];

	if (y == 1)
		return GetPredictionJPEGLSDe(x, y);

	if (x == 0)
		return R[0][y] - R[0][y - 1];

	if (x == 1)
		return GetPredictionJPEGLSDe(x, y);

	int rBorder = x1 + BSize - 2;
	if (rBorder >= W - 1)
		rBorder = W - 1;
	if (x >= rBorder)
		return GetPredictionJPEGLSDe(x, y);

	double Wa = R[x - 1][y];
	double NW = R[x - 1][y - 1];
	double N = R[x][y - 1];
	double NE = R[x + 1][y - 1];
	double NN = R[x][y - 2];
	double NNE = R[x + 1][y - 2];
	double WW = R[x - 2][y];

	double dh = abs(Wa - WW) + abs(N - NW) + abs(NE - N);
	double dv = abs(Wa - NW) + abs(N - NN) + abs(NE - NNE);


	if (dv - dh > 80) return (int)Wa;
	if (dv - dh < -80) return (int)N;

	double m = ((int)(Wa + N) / 2) + ((int)(NE - NW) / 4);

	if (dv - dh > 32) return ((int)(m + Wa) / 2);
	if (dv - dh > 8) return ((int)(3 * m + Wa) / 4);
	if (dv - dh < -32) return ((int)(m + N) / 2);
	if (dv - dh < -8) return ((int)(3 * m + N) / 4);

	return (int)m;
}


void MyDeRBlock::ReconstructCALIC(int x, int y)
{
	SetBlockCoordinates(x, y);

	
	for (int j = y1; j < y2; j++)
		for (int i = x1; i < x2; i++)
			R[i][j] = GetPredictionCALICDe(i, j) - UnWrapePositiveValues(P[i][j]);

}


vector<vector<int>> MyDeRBlock::RestorePixels()
{
	FreeMatrix(R);
	R = AllocateMatrix(W, H);
	ReconstructLUBorder();

	int k = 0;
	int Bcounter = 0;

	for (int x = 0; x < W; x += BSize)
	{
		for (int y = 0; y < H; y += BSize)
		{
			switch (ControlVector[k])
			{
//			case RAW_BLOCK: ReconstructRawBlock(x, y, ControlVector[++k]); break;
			case RAW_BLOCK: ReconstructRawBlock(x, y); break;
			case JPEGLS_BLOCK: ReconstructJPEGLS(x, y); break;
			case HORIZONTAL_BLOCK: ReconstructHorizontal(x, y); break;
			case VERTICAL_BLOCK: ReconstructVertical(x, y); break;
			case MTF_BLOCK: ReconstructMTF(x, y); break;
			case MWI_BLOCK: ReconstructMwI(x, y); break;
			case DIAGONAL_BLOCK: ReconstructDiagonal(x, y); break;
			case IDW_BLOCK: ReconstructIDW(x, y); break;
			case UPLEFT_BLOCK: ReconstructUpLeftAverage(x, y); break;
			case PAETH_BLOCK: ReconstructPeath(x, y); break;
			case CALIC_BLOCK: ReconstructCALIC(x, y); break;
			}
			k++;			
		}
	}

	return R;
}


