#include "MyRBlock.h"
#include "BlockMwITransform.h"

#include <math.h>
#include <stdio.h>
#include <string>
#include <algorithm>

#include "MyDefines.h"
#include myPATH(MY_INCLUDES_DIR,AuxCompFunctions.h)
#include myPATH(MY_INCLUDES_DIR,MTFWithIntegers.h)
#include myPATH(MY_INCLUDES_DIR,MwI.h)



#pragma warning (disable : 4996)

//#define PRINTOUT 0


void MyRBlock::Init(int X, int Y, int b, int MwId, double IDWd, vector <vector<int>> M)
{
	W = X;
	H = Y;
	I = M;
	BSize = b;
	MwIDelta = MwId;
	IDWDistance = IDWd;
	O = AllocateMatrix(W, H);
	PredictLUBorder();
	GE.resize(512);
	for (int i = 0; i < GE.size(); i++)
		GE[i] = 0;

	GEAux.resize(512);

	BCounters.resize(NO_OF_PREDICTORS);
	for (int i = 0; i < BCounters.size(); i++)
		BCounters[i] = 0;
}


int MyRBlock::Save(const char* FName)
{
	if (ControlVector.size() == 0)
		return 0;

	FILE* f;
	f = fopen(FName, "wt");

	fprintf(f, "%d %d\n", W, H);
	fprintf(f, "%d \n", BSize);
	fprintf(f, "%d \n", MwIDelta);
	fprintf(f, "%lf \n", IDWDistance);

	fprintf(f, "%d\n", (int)ControlVector.size());
	for (int i = 0; i < (int)ControlVector.size(); i++)
		fprintf(f, "%d ", ControlVector[i]);
	fprintf(f, "\n");

	for (int j = 0; j < H; j++)
	{
		for (int i = 0; i < W; i++)
		{
			fprintf(f, "%*d ", -3, O[i][j]);
		}
		fprintf(f, "\n");
	}

	fclose(f);

	return 1;
}


void MyRBlock::PredictLUBorder()
{
	O[0][0] = I[0][0];

	for (int x = 1; x < W; x++)
		O[x][0] =  ToPositiveValue(I[x][0] - I[x-1][0]);

	for (int y = 1; y < H; y++)
		O[0][y] = ToPositiveValue(I[0][y] - I[0][y-1]);
}


int  MyRBlock::PrintOToFile(const char* fName)
{

	return PrintMatrix(O, fName);

	return 1;

}

void  MyRBlock::Statistic(int BestMethod)
{

	BCounters[BestMethod]++;

}

void MyRBlock::GlobalEntropy()
{ 
	vector<vector<int>> A;
	vector<int> GEStorage;
	int x, y;
	double e;
	double emin = 16;
	int sum, summin; 
	summin = 1000000;
	for (int i = 1; i < NO_OF_PREDICTORS; i++)
	{
		GEAux = GE;
		switch (i)
		{
		case JPEGLS_BLOCK: A = JPEGLSBlock; break;
		case HORIZONTAL_BLOCK: A = HorizontalBlock; break;
		case VERTICAL_BLOCK: A = VerticalBlock; break;
		case MTF_BLOCK: A = MTFBlock; break;
		case MWI_BLOCK: A = MwIBlock; break;
		case DIAGONAL_BLOCK: A = DiagonalBlock; break;
		case IDW_BLOCK: A = IDWBlock; break;
		case UPLEFT_BLOCK: A = UpLeftBlock; break; 
		case PAETH_BLOCK: A = PaethBlock; break;
		case CALIC_BLOCK: A = CalicBlock; break;
		case AKIMA_BLOCK: A = AkimaBlock; break;
		}

		sum = 0;
		for (x = 0; x < (int)A.size(); x++)
			for (y = 0; y < (int)A[0].size(); y++)
			{
				GEAux[A[x][y]]++;
				sum += A[x][y];
			}
		
		e = CalculateEntropy(GEAux);

		if (sum < summin)
		{
			summin = sum;
			emin = e;
			GEStorage = GEAux;
		}
	}

	GE = GEStorage;

	printf("Emin %f \n", emin);
}

void MyRBlock::PrintAllPredictionBlocks()
{
	PrintMatrix(JPEGLSBlock, "JPEGLS.txt");
	PrintMatrix(HorizontalBlock, "HORIZONTAL.txt");
	PrintMatrix(VerticalBlock, "VERTICAL.txt");
	PrintMatrix(MTFBlock, "MTF.txt");
	PrintMatrix(MwIBlock, "MWI.txt");
	PrintMatrix(DiagonalBlock, "Diagonal.txt");
	PrintMatrix(IDWBlock, "IDW.txt");
	PrintMatrix(UpLeftBlock, "IDW.txt");
	PrintMatrix(PaethBlock, "Paeth.txt");
	PrintMatrix(CalicBlock, "Calic.txt");
	PrintMatrix(AkimaBlock, "Akima.txt");
}


void PrintStatisticInfo(const char* niz, int no, int total)
{
	double rat = (double)no / (double)total;
	rat *= 100;
	printf("%s, %5d, %6.2f%%\n", niz, no, rat);

}


vector<char>  MyRBlock::DetermineControlBits()
{
	vector<char> bts;
	vector<int> Z;

	int i;
	for (i = 0; i < NO_OF_PREDICTORS; i++)
		if (BCounters[i] != 0)
			Z.push_back(BCounters[i]);

	sort(Z.begin(), Z.end(), greater<int>());

	for (i = 0; i < Z[0]; i++)
		bts.push_back('0');
	
	for (i = 0; i < Z[1]; i++)
	{
		bts.push_back('1');
		bts.push_back('0');
	}
		
	for (i = 0; i < Z[2]; i++)
	{
		bts.push_back('1');
		bts.push_back('1');
	}



	return bts;
}


void  MyRBlock::PrintStatistic()
{

	printf("\n");

	int NoOfBlocks = 0;
	for (int i = 0; i < (int)BCounters.size(); i++)
		NoOfBlocks += BCounters[i];

	printf("NoOfBlocks: %d\n", NoOfBlocks);

	PrintStatisticInfo("B[0][0]    : ", BCounters[RAW_BLOCK], NoOfBlocks); 
	PrintStatisticInfo("a          : ", BCounters[VERTICAL_BLOCK], NoOfBlocks);
	PrintStatisticInfo("b          : ", BCounters[HORIZONTAL_BLOCK], NoOfBlocks);
	PrintStatisticInfo("B[i-1][j-1]: ", BCounters[DIAGONAL_BLOCK], NoOfBlocks);
	PrintStatisticInfo("(a+b)/2    : ", BCounters[UPLEFT_BLOCK], NoOfBlocks);
	PrintStatisticInfo("MTF        : ", BCounters[MTF_BLOCK], NoOfBlocks);
	PrintStatisticInfo("MwI(11)    : ", BCounters[MWI_BLOCK], NoOfBlocks);
	PrintStatisticInfo("IDW(9)     : ", BCounters[IDW_BLOCK], NoOfBlocks);
	PrintStatisticInfo("Paeth      : ", BCounters[PAETH_BLOCK], NoOfBlocks);
	PrintStatisticInfo("JPEGLS     : ", BCounters[JPEGLS_BLOCK], NoOfBlocks);
	PrintStatisticInfo("CALIC      : ", BCounters[CALIC_BLOCK], NoOfBlocks);

/*	for (int i = 0; i < (int)BCounters.size(); i++)
		switch (i)
		{
		case RAW_BLOCK       : PrintStatisticInfo("Constant Prediction:  ", BCounters[i], NoOfBlocks); break;
		case HORIZONTAL_BLOCK: PrintStatisticInfo("Horizontal Prediction:", BCounters[i], NoOfBlocks); break;
		case VERTICAL_BLOCK  : PrintStatisticInfo("Vertical Prediction:  ", BCounters[i], NoOfBlocks); break;
		case DIAGONAL_BLOCK  : PrintStatisticInfo("Diagonal Prediction:  ", BCounters[i], NoOfBlocks); break;
		case UPLEFT_BLOCK    : PrintStatisticInfo("Up-left Prediction :  ", BCounters[i], NoOfBlocks); break;
		case MTF_BLOCK       : PrintStatisticInfo("MTF Prediction:       ", BCounters[i], NoOfBlocks); break;
		case MWI_BLOCK       : PrintStatisticInfo("MwI Prediction:       ", BCounters[i], NoOfBlocks); break;
		case IDW_BLOCK       : PrintStatisticInfo("IDW Prediction:       ", BCounters[i], NoOfBlocks); break;
		case PAETH_BLOCK     : PrintStatisticInfo("Paeth Prediction:     ", BCounters[i], NoOfBlocks); break;
		case JPEGLS_BLOCK    : PrintStatisticInfo("JPEGLS Prediction:    ", BCounters[i], NoOfBlocks); break;
		case CALIC_BLOCK	 : PrintStatisticInfo("CALIC Prediction:     ", BCounters[i], NoOfBlocks); break;
//		case AKIMA_BLOCK     : PrintStatisticInfo("Akima Prediction:     ", BCounters[i], NoOfBlocks); break;
		}
		*/
}



void MyRBlock::PrintStatisticRow()
{
	printf("\n");

	int NoOfBlocks = 0;
	for (int i = 0; i < (int)BCounters.size(); i++)
		NoOfBlocks += BCounters[i];

	printf("NoOfBlocks: %d\n", NoOfBlocks);
	printf("|%7d  ", BCounters[RAW_BLOCK]);
	printf("|%6d   ", BCounters[VERTICAL_BLOCK]);
	printf("|%6d   ", BCounters[HORIZONTAL_BLOCK]);
	printf("|%7d  ", BCounters[DIAGONAL_BLOCK]);
	printf("|%7d  ", BCounters[UPLEFT_BLOCK]);
	printf("|%5d  ", BCounters[MTF_BLOCK]);
	printf("|%7d  ", BCounters[MWI_BLOCK]);
	printf("|%5d  ", BCounters[IDW_BLOCK]);
	printf("|%5d  ", BCounters[PAETH_BLOCK]);
	printf("|%6d  ", BCounters[JPEGLS_BLOCK]);
	printf("|%6d  ", BCounters[CALIC_BLOCK]);
	printf("\n");

	printf("percentages \n");
	printf("| %6.3f  ", 100 * (double)BCounters[RAW_BLOCK] / (double)NoOfBlocks);
	printf("| %6.3f  ", 100 * (double)BCounters[VERTICAL_BLOCK] / (double)NoOfBlocks);
	printf("| %6.3f  ", 100 * (double)BCounters[HORIZONTAL_BLOCK] / (double)NoOfBlocks);
	printf("| %8.3f ", 100 * (double)BCounters[DIAGONAL_BLOCK] / (double)NoOfBlocks);
	printf("| %6.3f  ", 100 * (double)BCounters[UPLEFT_BLOCK] / (double)NoOfBlocks);
	printf("| %5.3f ", 100 * (double)BCounters[MTF_BLOCK] / (double)NoOfBlocks);
	printf("| %5.3f ", 100 * (double)BCounters[MWI_BLOCK] / (double)NoOfBlocks);
	printf("| %5.3f ", 100 * (double)BCounters[IDW_BLOCK] / (double)NoOfBlocks);
	printf("| %5.3f ", 100 * (double)BCounters[PAETH_BLOCK] / (double)NoOfBlocks);
	printf("| %6.3f ", 100 * (double)BCounters[JPEGLS_BLOCK] / (double)NoOfBlocks);
	printf("| %6.3f ", 100 * (double)BCounters[CALIC_BLOCK] / (double)NoOfBlocks);
	printf("\n");

}



double MyRBlock::FinalEntropy()
{
	return CalculateEntropy(DoMatrixStatistic(H, W, O));
}

void MyRBlock::SetBlockCoordinates(int x, int y)
{
	if (x == 0)
		x1 = x + 1;
	else
		x1 = x;

	if (y == 0)
		y1 = y + 1;
	else
		y1 = y;

	x2 = x + BSize;
	if (x2 > W)
		x2 = W;

	y2 = y + BSize;
	if (y2 > H)
		y2 = H;

	dx = x2 - x1;
	dy = y2 - y1;

}



void MyRBlock::SetBlock(int x, int y)
{
	SetBlockCoordinates(x, y);
	FreeMatrix(Block);
	Block = AllocateMatrix(dx, dy);
	
	int a, b;
	a = 0;
	for (int i = x1; i < x2; i++)
	{
		b = 0;
		for (int j = y1; j < y2; j++)
			Block[a][b++] = I[i][j];
		a++;
	}
#ifdef PRINTOUT
	PrintMatrix(Block, "BLokcek.txt");
#endif // PRINTOUT

	
}

void MyRBlock::StoreBest(int bestIndex)
{
	FreeMatrix(ABlock);
	ABlock = AllocateMatrix(dx, dy);
	if (bestIndex == RAW_BLOCK)
		ABlock = LowValueRawBlock;

	if (bestIndex == JPEGLS_BLOCK)
		ABlock = JPEGLSBlock;

	if (bestIndex == HORIZONTAL_BLOCK)
		ABlock = HorizontalBlock;

	if (bestIndex == VERTICAL_BLOCK)
		ABlock = VerticalBlock;

	if (bestIndex == MTF_BLOCK)
		ABlock = MTFBlock;

	if (bestIndex == MWI_BLOCK)
		ABlock = MwIBlock;

	if (bestIndex == DIAGONAL_BLOCK)
		ABlock = DiagonalBlock;

	if (bestIndex == IDW_BLOCK)
		ABlock = IDWBlock;

	if (bestIndex == UPLEFT_BLOCK)
		ABlock = UpLeftBlock;

	if (bestIndex == PAETH_BLOCK)
		ABlock = PaethBlock;

	if (bestIndex == CALIC_BLOCK)
		ABlock = CalicBlock;

	if (bestIndex == AKIMA_BLOCK)
		ABlock = AkimaBlock;


	ControlVector.push_back(bestIndex);

	int i, j;
	i = 0;
	for (int x = x1; x < x2; x++)
	{
		j = 0;
		for (int y = y1; y < y2; y++)
			O[x][y] = ABlock[i][j++];
		i++;
	}

#ifdef PRINTOUT
	PrintMatrix(O, "O.txt");
#endif // PRINTPUT


	
}

/*
void MyRBlock::RawPrediction()
{
	FreeMatrix(LowValueRawBlock);
	LowValueRawBlock = AllocateMatrix(dx, dy);
	LowValue = 256;

	for (int x = 0; x < dx; x++)
		for (int y = 0; y < dy; y++)
			if (Block[x][y] < LowValue)
				LowValue = Block[x][y];

	for (int x = 0; x < dx; x++)
		for (int y = 0; y < dy; y++)
			LowValueRawBlock[x][y] = Block[x][y] - LowValue;
}*/

void MyRBlock::RawPrediction()
{

	FreeMatrix(PBlock);
	PBlock = AllocateMatrix(dx, dy);

	int i, j;
	i = 0;
	int leftup = I[x1 - 1][y1 - 1];
	for (int x = x1; x < x2; x++)
	{
		j = 0;
		for (int y = y1; y < y2; y++)
			PBlock[i][j++] = I[x][y] - leftup;
		i++;
	}


	FreeMatrix(LowValueRawBlock);
	LowValueRawBlock = AllocateMatrix(dx, dy);

	for (int x = 0; x < dx; x++)
		for (int y = 0; y < dy; y++)
			LowValueRawBlock[x][y] = ToPositiveValue(PBlock[x][y]);
}


double MyRBlock::GetRawBlockEntropy()
{
	FreeMatrix(LowValueRawBlock);
	LowValueRawBlock = AllocateMatrix(dx, dy);
	LowValue = 256;
	int HighValue = -1;
	for (int x = 0; x < dx; x++)
		for (int y = 0; y < dy; y++)
		{
			if (Block[x][y] < LowValue)
				LowValue = Block[x][y];

			if (Block[x][y] > HighValue)
				HighValue = Block[x][y];
		}

//	if (HighValue - LowValue > 15)
//		return 255;

	for (int x = 0; x < dx; x++)
		for (int y = 0; y < dy; y++)
			LowValueRawBlock[x][y] = Block[x][y] - LowValue;

	return CalculateEntropy(DoMatrixStatistic((int)LowValueRawBlock[0].size(), (int)LowValueRawBlock.size(), LowValueRawBlock));
}



void MyRBlock::JPEGLJPrediction()
{
	FreeMatrix(PBlock);
	PBlock = AllocateMatrix(dx, dy);


	int i, j;
	i =  0;
	for (int x = x1; x < x2; x++)
	{
		j = 0;
		for (int y = y1; y < y2; y++)
		{
			PBlock[i][j] = GetPredictionJPEGLS(x, y) - I[x][y];
			j++;
		}
		i++;
	}

	FreeMatrix(JPEGLSBlock);
	JPEGLSBlock = AllocateMatrix(dx, dy);

	for (int y = 0; y < dy; y++)
		for (int x = 0; x < dx; x++)
			JPEGLSBlock[x][y] = ToPositiveValue(PBlock[x][y]);
}

int MyRBlock::GetPredictionJPEGLS(int x, int y)
{
	if (x == 0)
		return I[x][y - 1];

	int a = I[x][y - 1];
	int c = I[x - 1][y - 1];
	int b = I[x - 1][y];

	if ((c > a) && (c > b))
		if (a < b)
			return a;
		else
			return b;

	if ((c < a) && (c < b))
		if (a > b)
			return a;
		else
			return b;

	int d = a + b - c;
	return d;
}

void MyRBlock::HorizontalPrediction()
{
	FreeMatrix(PBlock);
	PBlock = AllocateMatrix(dx, dy);
	int i, j;
	j = 0;
	for (int y = y1; y < y2; y++)
	{
		i = 0;
		for (int x = x1; x < x2; x++)
		{
			PBlock[i][j] = I[x][y - 1] - I[x][y];
			i++;
		}
		j++;
	}

	FreeMatrix(HorizontalBlock);
	HorizontalBlock = AllocateMatrix(dx, dy);
	for (int y = 0; y < dy; y++)
		for (int x = 0; x < dx; x++)
			HorizontalBlock[x][y] = ToPositiveValue(PBlock[x][y]);

}


void MyRBlock::VerticalPrediction()
{
	FreeMatrix(PBlock);
	PBlock = AllocateMatrix(dx, dy);
	int i, j;
	j = 0;
	for (int y = y1; y < y2; y++)
	{
		i = 0;
		for (int x = x1; x < x2; x++)
		{
			PBlock[i][j] = I[x-1][y] - I[x][y];
			i++;
		}
		j++;
	}

	FreeMatrix(VerticalBlock);
	VerticalBlock = AllocateMatrix(dx, dy);

	for (int y = 0; y < dy; y++)
		for (int x = 0; x < dx; x++)
			VerticalBlock[x][y] = ToPositiveValue(PBlock[x][y]);

}


vector<int> MyRBlock::FillBlockBorderValues()
{
	vector<int> V;

	int x = x1 - 1;
	int y = y1 - 1;

	V.push_back(I[x][y]);

	for (int i = 0; i < dx + dy; i++)
	{
		x++;
		if (x < x2)
			V.push_back(I[x][y1 - 1]);

		y++;
		if (y < y2)
			V.push_back(I[x1 - 1][y]);
	}

	return V;
}

void MyRBlock::MTF()
{
	FreeMatrix(MTFBlock);
	MTFBlock = AllocateMatrix(dx, dy);

	vector<int> BV = FillBlockBorderValues();
	MTFWithIntegers mtf;
	mtf.Init(0, 255, BV);

	vector<int> BlockVector;
	for (int y = y1; y < y2; y++)
		for (int x = x1; x < x2; x++)
			BlockVector.push_back(I[x][y]);
	
	vector<int> MTFOut = mtf.DoMTF(BlockVector);

	int k = 0;
	for (int y = 0; y < dy; y++)
		for (int x = 0; x < dx; x++)
			MTFBlock[x][y] = MTFOut[k++];

}

void MyRBlock::BlockMwI()
{
	FreeMatrix(MwIBlock);
	MwIBlock = AllocateMatrix(dx, dy);

	BlockMwITransform BMwIT;
	BMwIT.Init(0, 255, MwIDelta);

	int i, j;
	i = 0;
	int a, b, c, d;

	for (int x = x1; x < x2; x++)
	{
		j = 0;
		for (int y = y1; y < y2; y++)
		{

			a = I[x - 1][y];
			b = I[x - 1][y - 1];
			c = I[x][y - 1];
			d = -1;
			if (x + 1 < x2)
				d = I[x + 1][y - 1];


			MwIBlock[i][j] = BMwIT.TransformMwI(I[x][y], a, b, c, d);
			j++;
		}
		i++;
	}
}



void MyRBlock::Diagonal()
{	

	FreeMatrix(PBlock);
	PBlock = AllocateMatrix(dx, dy);
	int i, j;
	j = 0;
	int xd, yd;
	for (int x = x1; x < x2; x++)
	{
		xd = x;
		yd = y1;
		i = xd - x1;
		j = 0;
		while ((xd < x2) && (yd < y2))
		{
			PBlock[i][j] = I[xd - 1][yd - 1] - I[xd][yd];
			xd++;
			yd++;
			i++;
			j++;
		}
	}
	
	for (int y = y1+1; y < y2; y++)
	{
		xd = x1;
		yd = y;
		j = yd - y1;
		i = 0;
		while ((xd < x2) && (yd < y2))
		{
			PBlock[i][j] = I[xd - 1][yd - 1] - I[xd][yd];
			xd++;
			yd++;
			i++;
			j++;
		}
	}

	FreeMatrix(DiagonalBlock);
	DiagonalBlock = AllocateMatrix(dx, dy);

	for (int y = 0; y < dy; y++)
		for (int x = 0; x < dx; x++)
			DiagonalBlock[x][y] = ToPositiveValue(PBlock[x][y]);
}

void MyRBlock::IDW()
{
	FreeMatrix(PBlock);
	PBlock = AllocateMatrix(dx, dy);

	vector<int> X, Y, V;

	int xx = x1 - 1;
	int yy = y1 - 1;
	for (int x = xx; x < x2; x++)
	{
		X.push_back(x); 
		Y.push_back(yy);
		V.push_back(I[x][yy]);
	}

	for (int y = yy; y < y2; y++)
	{
		X.push_back(xx);
		Y.push_back(y);
		V.push_back(I[xx][y]);
	}

	int i, j;
	int v;
	j = 0;
	for (int y = y1; y < y2; y++)
	{
		i = 0;
		for (int x = x1; x < x2; x++)
		{
			v = ReturnIDW(X, Y, V, x, y, IDWDistance);
			PBlock[i][j] = v - I[x][y];
			i++;
			X.push_back(x);
			Y.push_back(y);
			V.push_back(I[x][y]);
		}
		j++;
	}

	FreeMatrix(IDWBlock);
	IDWBlock = AllocateMatrix(dx, dy);

	for (int y = 0; y < dy; y++)
		for (int x = 0; x < dx; x++)
			IDWBlock[x][y] = ToPositiveValue(PBlock[x][y]);

}

void MyRBlock::UpLeftAverage()
{

	FreeMatrix(PBlock);
	PBlock = AllocateMatrix(dx, dy);
	int i, j;
	j = 0;
	for (int y = y1; y < y2; y++)
	{
		i = 0;
		for (int x = x1; x < x2; x++)
		{
			PBlock[i][j] = (int)(0.5 * (I[x - 1][y] + I[x][y - 1])) - I[x][y];
			i++;
		}
		j++;
	}

	FreeMatrix(UpLeftBlock);
	UpLeftBlock = AllocateMatrix(dx, dy);

	for (int y = 0; y < dy; y++)
		for (int x = 0; x < dx; x++)
			UpLeftBlock[x][y] = ToPositiveValue(PBlock[x][y]);
}


int PaethPrediction(int a, int b, int c)
// a: left, b: up, c: left up pixel
{
	int p = a + b - c;
	int pa = abs(p - a);
	int pb = abs(p - b);
	int pc = abs(p - c);
	if ((pa <= pb) && (pa <= pc)) return a;
	if (pb <= pc) return b;
	return c;
}

void MyRBlock::Paeth()
{
	FreeMatrix(PBlock);
	PBlock = AllocateMatrix(dx, dy);
	int i, j;
	j = 0;
	for (int y = y1; y < y2; y++)
	{
		i = 0;
		for (int x = x1; x < x2; x++)
		{
			PBlock[i][j] = PaethPrediction(I[x-1][y], I[x][y-1], I[x-1][y-1]) - I[x][y];
			i++;
		}
		j++;
	}

	FreeMatrix(PaethBlock);
	PaethBlock = AllocateMatrix(dx, dy);

	for (int y = 0; y < dy; y++)
		for (int x = 0; x < dx; x++)
			PaethBlock[x][y] = ToPositiveValue(PBlock[x][y]);
}


int MyRBlock::GetPredictionCALIC(int x, int y)
{

	if ((x == 0) && (y == 0))
		return I[0][0];

	if (y == 0)
		return I[x][0] - I[x - 1][0];

	if (y == 1)
		return GetPredictionJPEGLS(x, y);

	if (x == 0)
		return I[0][y] - I[0][y - 1];

	if (x == 1)
		return GetPredictionJPEGLS(x, y);

	int rBorder = x1 + BSize - 2;
	if (rBorder >= W - 1)
		rBorder = W - 1;
	if (x >= rBorder)
		return GetPredictionJPEGLS(x, y);

	double Wa = I[x - 1][y];
	double NW = I[x - 1][y - 1];
	double N = I[x][y-1];
	double NE = I[x + 1][y - 1];
	double NN = I[x][y - 2];
	double NNE = I[x + 1][y - 2];
	double WW = I[x - 2][y];

	double dh = abs(Wa - WW) + abs(N - NW) + abs(NE - N);
	double dv = abs(Wa - NW) + abs(N - NN) + abs(NE - NNE);


	if (dv - dh > 80) return (int)Wa;
	if (dv - dh < -80) return (int)N;

	double m = ((int)(Wa + N) / 2) + ((int)(NE - NW) / 4);

	if (dv - dh > 32) return ((int)(m + Wa) / 2);
	if (dv - dh > 8) return ((int)(3 * m + Wa) / 4);
	if (dv - dh < -32) return ((int)(m + N) / 2);
	if (dv - dh < -8) return ((int)(3 * m + N) / 4);

	return (int)m;
}


void MyRBlock::Calic()
{
	FreeMatrix(PBlock);
	PBlock = AllocateMatrix(dx, dy);
	int i, j;
	j = 0;
	for (int y = y1; y < y2; y++)
	{
		i = 0;
		for (int x = x1; x < x2; x++)
		{
			PBlock[i][j] = GetPredictionCALIC(x, y) - I[x][y];
			i++;
		}
		j++;
	}

	FreeMatrix(CalicBlock);
	CalicBlock = AllocateMatrix(dx, dy);

	for (int y = 0; y < dy; y++)
		for (int x = 0; x < dx; x++)
			CalicBlock[x][y] = ToPositiveValue(PBlock[x][y]);
}


void MyRBlock::SetAkimaControlPoints(int yy, vector<int>& X,  vector<int>& V)
{
	X.clear();
	V.clear();

	int xx = x1 - 1;
	vector<int> X1, V1;
	X1.push_back(xx);

	V1.push_back(I[xx][yy]);

	int q1 = xx + (int)(BSize * 0.333);
	int q2 = xx + (int)(BSize * 0.666);
	
	int p1, p2;
	p1 = p2 = x1;
	int d1, d2;
	d1 = d2 = 0;

	for (int i = x1; i < x2; i++) // find first bigest difference
		if (abs(I[i][yy] - I[i - 1][yy]) > d1)
		{
			d1 = abs(I[i][yy] - I[i - 1][yy]);
			p1 = i;
		}

	for (int i = x1; i < x2; i++) // find second bigest difference
		if (i != p1)
			if (abs(I[i][yy] - I[i - 1][yy]) > d2)
			{
				d2 = abs(I[i][yy] - I[i - 1][yy]);
				p2 = i;
			}

	q1 = p1;
	q2 = p2;
	if (q1 > q2)
	{
		int a = q1;
		q1 = q2;
		q2 = a;
	}

	X1.push_back(q1); V1.push_back(I[q1][yy]);
	X1.push_back(q2); V1.push_back(I[q2][yy]);

	X1.push_back(x2-1); V1.push_back(I[x2-1][yy]);

	X.push_back(X1[0]); V.push_back(V1[0]);
	for (int i = 1; i < X1.size(); i++)
	{
		if (X1[i - 1] != X1[i])
		{
			X.push_back(X1[i]);
			V.push_back(V1[i]);
		}
	}
}

void MyRBlock::InterpolateAkima(vector<int>& X, vector<int>& V, vector<int>& O)
{
	double a;

	O.clear();
	for (int k = 0; k < X.size() - 1; k++)
	{
		int b1 = X[k];
		int b2 = X[k + 1];
		if (b2 == b1 + 1)
		{
			O.push_back(V[k]);
			O.push_back(V[k+1]);
		}
		else
		{
			double m = (double)(V[k + 1] - V[k]) / (double)(X[k + 1] - X[k]);
			double w, d;
			if (k == 0)
				w  = 1 / (double)abs(X[k + 1] - X[k]);
			else
			{
				if (k == (int)X.size() - 1)
					w = 1 / (double)abs(X[k] - X[k - 1]);
				else
				{
					double r = abs(X[k + 1] - X[k]) + abs(X[k] - X[k - 1]);
					w = 1 / r;
				}
			}
			for (int x = b1+1; x < b2; x++)
			{
				d = x - b1;
				a = V[k] + m * d + w * 0.5 * d * d - w * 0.5 * d * d * d;
				int z = (int)round(a);
				O.push_back(z);
			}		
			O.push_back(V[k+1]);
		}
	}

}


void MyRBlock::Akima()
{
	FreeMatrix(PBlock);
	PBlock = AllocateMatrix(dx, dy);

	FreeMatrix(AkimaBlock);
	AkimaBlock = AllocateMatrix(dx, dy);


	vector<int> X, V;

	int xx = x1 - 1;
	int yy = y1 - 1;
	int LBorder = xx + BSize;
	int BBorder = yy + BSize;
	if ((LBorder >= W) || (BBorder >= H))
		InitAllocatedMatrix(AkimaBlock, dx, dy, EMPTYVALUE);
	else
	{
		vector<int> X, V;
		vector<int> AI;
		SetAkimaControlPoints(yy, X,V);

		int yyb = 0;
		for (int y = y1; y < y1+dy; y++)
		{
			InterpolateAkima(X, V, AI);
			int xx1 = x1;

			for (int i = 1; i < (int)AI.size()-1; i++)
			{				
				PBlock[i-1][yyb] = AI[i] - I[xx1][y];
				xx1++;
			}
			yyb++;
			SetAkimaControlPoints(y, X, V);
		}

		for (int y = 0; y < dy; y++)
			for (int x = 0; x < dx; x++)
				AkimaBlock[x][y] = ToPositiveValue(PBlock[x][y]);
	}
}



