#pragma once
#include <vector>
#include "MyDefines.h"
#include myPATH(MY_INCLUDES_DIR,AuxCompFunctions.h)

#include "MyRBlockDefs.h"

using namespace std;


class MyRBlock
{ 
private:
	int W, H;
	int BSize;
	int MwIDelta;
	double IDWDistance;

	int x1, y1, x2, y2; // Block coordiante
	int dx, dy; // the size od the actual block
	
	vector<vector<int>> O; // resulting prediction
	vector<int> ControlVector;
	vector<int> GE; // Glocal Entropy
	vector<int> GEAux;
	vector<int> BCounters; // for statistics
	

	vector<vector<int>> I;
	vector<vector<int>> Block;
	vector<vector<int>> PBlock; // prediction block
	vector<vector<int>> ABlock; // auxiliary block

	vector<vector<int>> LowValueRawBlock;
	vector<vector<int>> JPEGLSBlock;
	vector<vector<int>> HorizontalBlock;
	vector<vector<int>> VerticalBlock;
	vector<vector<int>> MTFBlock;
	vector<vector<int>> MwIBlock;
	vector<vector<int>> DiagonalBlock;
	vector<vector<int>> IDWBlock;
	vector<vector<int>> UpLeftBlock;
	vector<vector<int>> PaethBlock;
	vector<vector<int>> CalicBlock;
	vector<vector<int>> AkimaBlock;


	vector<vector<int>> R; // reconstructed

	void SetBlockCoordinates(int x, int y);

	int LowValue;



	void PredictLUBorder();

	int GetPredictionJPEGLS(int x, int y);
	int GetPredictionCALIC(int x, int y);

	vector<int> FillBlockBorderValues();

	void SetAkimaControlPoints(int y, vector<int>& X, vector<int>& V);
	void InterpolateAkima(vector<int>& X, vector<int>& V, vector<int>& O);

public:
	MyRBlock() {};

	void Init(int X, int Y, int BlockSize, int MwIDelta, double IDWd, vector<vector<int>> M);

	int Save(const char*);

	void SetBlock(int x, int y);
	void StoreBest(int bestIndex);
	void Statistic(int BestMethod);
	void PrintStatistic();
	void PrintStatisticRow();
	void PrintAllPredictionBlocks();
	double FinalEntropy();

	void GlobalEntropy();

	void RawPrediction();
	double GetRawBlockEntropy();
	double GetRawStandardDeviation() { return CalculateStandardDeviation(LowValueRawBlock); }


	void JPEGLJPrediction();
	double GetJPEGLSStandardDeviation() { return CalculateStandardDeviation(JPEGLSBlock); }

	void HorizontalPrediction();
	double GetHorizontalStandardDeviation() { return CalculateStandardDeviation(HorizontalBlock); }

	void VerticalPrediction();
	double GetVerticalStandardDeviation() { return CalculateStandardDeviation(VerticalBlock); }

	void MTF();
	double GetMTFStandardDeviation() { return CalculateStandardDeviation(MTFBlock); }

	void BlockMwI();
	double GetMwIStandardDeviation() { return CalculateStandardDeviation(MwIBlock); }

	void Diagonal();
	void IDW();
	void UpLeftAverage();
	void Paeth();
	void Calic();
	void Akima();


	int GetRawBlockSum() { return SumMatrixElements(LowValueRawBlock); }
	int GetJPEGLSBlockSum() { return SumMatrixElements(JPEGLSBlock); }
	int GetHorizontalBlockSum() { return SumMatrixElements(HorizontalBlock); }
	int GetVerticalBlockSum() { return SumMatrixElements(VerticalBlock); }
	int GetMTFBlockSum() { return SumMatrixElements(MTFBlock); }
	int GetMwIBlockSum() { return SumMatrixElements(MwIBlock); }
	int GetDiagonalBlockSum() { return SumMatrixElements(DiagonalBlock); }
	int GetIDWSum() { return SumMatrixElements(IDWBlock); }
	int GetUpLeftAverageSum() { return SumMatrixElements(UpLeftBlock); }
	int GetPaethSum() { return SumMatrixElements(PaethBlock); }
	int GetCalicSum() { return SumMatrixElements(CalicBlock); }
	int GetAkimaSum() { return SumMatrixElements(AkimaBlock); }



	int PrintOToFile(const char* fName);

	vector<char> DetermineControlBits();

};

